/*
 * This file is part of the Generic Data Structures Library (GDSL).
 * Copyright (C) 1998-2004 Nicolas Darnis <ndarnis@free.fr>.
 *
 * The GDSL library is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License as 
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * The GDSL library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the GDSL library; see the file COPYING.
 * If not, write to the Free Software Foundation, Inc., 
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
 *
 * $RCSfile: gdsl_queue.c,v $
 * $Revision: 1.16 $
 * $Date: 2004/09/30 11:00:48 $
 */

#include <config.h>


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>


#include "_gdsl_list.h"
#include "_gdsl_node.h"
#include "gdsl_types.h"
#include "gdsl_queue.h"


struct _gdsl_queue
{
    _gdsl_node_t      d;    /* begin of the queue (sentinel) */
    _gdsl_node_t      z;    /* end of the queue (sentinel) */
    char*             name; /* name of the queue */
    ulong             card; /* Cardinality of the queue */

    gdsl_alloc_func_t alloc_f;
    gdsl_free_func_t  free_f;
};

static gdsl_element_t 
default_alloc (void* e);

static void 
default_free (gdsl_element_t e);

static _gdsl_node_t
search_by_value (gdsl_queue_t q, gdsl_compare_func_t f, void* value);

static _gdsl_node_t 
search_by_position (gdsl_queue_t l, ulong pos);

/******************************************************************************/
/* Management functions of queues                                             */
/******************************************************************************/

extern gdsl_queue_t
gdsl_queue_alloc (const char* name, gdsl_alloc_func_t alloc_f, 
		  gdsl_free_func_t free_f)
{
    gdsl_queue_t q = NULL;

    q = (gdsl_queue_t) malloc (sizeof (struct _gdsl_queue));

    if (q == NULL)
	{
	    return NULL;
	}

    q->d = _gdsl_node_alloc ();

    if (q->d == NULL)
	{
	    free (q);
	    return NULL;
	}

    q->z = _gdsl_node_alloc ();

    if (q->z == NULL)
	{
	    _gdsl_node_free (q->d);
	    free (q);
	    return NULL;
	}

    q->name = NULL;

    if (gdsl_queue_set_name (q, name) == NULL)
	{
	    _gdsl_node_free (q->z);
	    _gdsl_node_free (q->d);
	    free (q);
	    return NULL;
	}

    q->card = 0UL;

    _gdsl_node_link (q->d, q->z);
    _gdsl_node_set_succ (q->z, q->z);
    _gdsl_node_set_pred (q->d, q->d);

    q->alloc_f = alloc_f ? alloc_f : default_alloc;
    q->free_f  = free_f  ? free_f  : default_free;

    return q;
}

extern void 
gdsl_queue_free (gdsl_queue_t q)
{
    assert (q != NULL);

    if (gdsl_queue_is_empty (q) == FALSE)
	{
	    gdsl_queue_flush (q);
	}

    _gdsl_node_free (q->d);
    _gdsl_node_free (q->z);

    if (q->name != NULL)
	{
	    free (q->name);
	}

    free (q);
}

extern void
gdsl_queue_flush (gdsl_queue_t q)
{
    _gdsl_node_t save;
    _gdsl_node_t tmp;

    assert (q != NULL);

    tmp = _gdsl_node_get_succ (q->d);

    while (tmp != q->z)
	{
	    save = _gdsl_node_get_succ (tmp);
	    q->free_f (_gdsl_node_get_content (tmp));
	    _gdsl_node_free (tmp);
	    tmp = save;
	}

    q->card = 0UL;

    _gdsl_node_link (q->d, q->z);
    _gdsl_node_set_succ (q->z, q->z);
    _gdsl_node_set_pred (q->d, q->d);
}

/******************************************************************************/
/* Consultation functions of queues                                           */
/******************************************************************************/

extern const char*
gdsl_queue_get_name (const gdsl_queue_t q)
{
    assert (q != NULL);

    return q->name;
}

extern ulong
gdsl_queue_get_size (const gdsl_queue_t q)
{
    assert (q != NULL);

    return q->card;
}

extern bool 
gdsl_queue_is_empty (const gdsl_queue_t q)
{
    assert (q != NULL);

    return (bool) (_gdsl_node_get_succ (q->d) == q->z);
}

extern gdsl_element_t
gdsl_queue_get_head (const gdsl_queue_t q)
{
    assert (q != NULL);

    return _gdsl_node_get_content (_gdsl_node_get_pred (q->z));
}

extern gdsl_element_t
gdsl_queue_get_tail (const gdsl_queue_t q)
{
    assert (q != NULL);

    return _gdsl_node_get_content (_gdsl_node_get_succ (q->d));
}

/******************************************************************************/
/* Modification functions of queues                                           */
/******************************************************************************/

extern gdsl_queue_t
gdsl_queue_set_name (gdsl_queue_t q, const char* name)
{
    assert (q != NULL);

    if (q->name != NULL)
	{
	    free (q->name);
	    q->name = NULL;
	}
  
    if (name != NULL)
	{
	    q->name = (char*) malloc ((1 + strlen (name)) * sizeof (char));

	    if (q->name == NULL)
		{
		    return NULL;
		}
    
	    strcpy (q->name, name);
	}

    return q;
}

extern gdsl_element_t
gdsl_queue_insert (gdsl_queue_t q, void* v)
{
    gdsl_element_t    e;
    _gdsl_node_t head = NULL;

    assert (q != NULL);

    head = _gdsl_node_alloc ();

    if (head == NULL)
	{
	    return NULL;
	}

    e = (q->alloc_f) (v);

    if (e == NULL)
	{
	    _gdsl_node_free (head);
	    return NULL;
	}

    q->card++;
    _gdsl_node_set_content (head, e);
  
    {
	_gdsl_node_t tmp = _gdsl_node_get_succ (q->d);
	_gdsl_node_link (q->d, head);
	_gdsl_node_link (head, tmp);
    }
  
    return e;
}

extern gdsl_element_t
gdsl_queue_remove (gdsl_queue_t q)
{
    assert (q != NULL);

    if (!gdsl_queue_is_empty (q))
	{
	    _gdsl_node_t tail = _gdsl_node_get_pred (q->z);
	    gdsl_element_t e = _gdsl_node_get_content (tail);

	    _gdsl_list_remove (tail);
	    _gdsl_node_free (tail);

	    q->card--;
	    return e;
	}

    return NULL;
}

/******************************************************************************/
/* Search functions of queues                                                 */
/******************************************************************************/

extern gdsl_element_t
gdsl_queue_search (const gdsl_queue_t q, gdsl_compare_func_t f, void* value)
{
    _gdsl_node_t n;

    assert (q != NULL);
    assert (f != NULL);

    n = search_by_value (q, f, value);

    return (n == NULL) ? NULL : _gdsl_node_get_content (n);
}

extern gdsl_element_t
gdsl_queue_search_by_position (const gdsl_queue_t q, ulong pos)
{
    _gdsl_node_t n;

    assert (q != NULL);
    assert (pos > 0 && pos <= q->card);

    n = search_by_position (q, pos);

    return n ? _gdsl_node_get_content (n) : NULL;
}

/******************************************************************************/
/* Parse functions of queues                                                  */
/******************************************************************************/

extern gdsl_element_t 
gdsl_queue_map_forward (const gdsl_queue_t q, gdsl_map_func_t f, void* user_data)
{
    gdsl_element_t e;
    _gdsl_node_t   tmp;

    assert (q != NULL);
    assert (f != NULL);

    tmp = _gdsl_node_get_succ (q->d);

    while (tmp != q->z)
	{
	    e = _gdsl_node_get_content (tmp);

	    if (f (e, user_data) == GDSL_MAP_STOP)
		{
		    return e;
		}

	    tmp = _gdsl_node_get_succ (tmp);
	}

    return NULL;
}

extern gdsl_element_t 
gdsl_queue_map_backward (const gdsl_queue_t q, gdsl_map_func_t f, void *user_data)
{
    gdsl_element_t e;
    _gdsl_node_t   tmp;

    assert (q != NULL);
    assert (f != NULL);

    tmp = _gdsl_node_get_pred (q->z);

    while (tmp != q->d)
	{
	    e = _gdsl_node_get_content (tmp);

	    if (f (e, user_data) == GDSL_MAP_STOP)
		{
		    return e;
		}

	    tmp = _gdsl_node_get_pred (tmp);
	}

    return NULL;
}

/******************************************************************************/
/* Input/output functions of queues                                           */
/******************************************************************************/

extern void
gdsl_queue_write (const gdsl_queue_t q, gdsl_write_func_t write_f, FILE *file, 
		  void *user_data)
{
    _gdsl_node_t tmp;

    assert (q != NULL);
    assert (write_f != NULL);
    assert (file != NULL);

    tmp = _gdsl_node_get_succ (q->d);

    while (tmp != q->z)
	{
	    _gdsl_node_write (tmp, write_f, file, user_data);
	    tmp = _gdsl_node_get_succ (tmp);
	}
}

extern void
gdsl_queue_write_xml (const gdsl_queue_t q, gdsl_write_func_t write_f, FILE *file, 
		      void *user_data)
{
    _gdsl_node_t tmp;

    assert (q != NULL);
    assert (file != NULL);

    tmp = _gdsl_node_get_succ (q->d);

    fprintf (file, "<GDSL_QUEUE REF=\"%p\" NAME=\"%s\" CARD=\"%ld\" HEAD=\"%p\" TAIL=\"%p\">\n", 
	     (void *) q, q->name, q->card, (void *) tmp, 
	     (void *) _gdsl_node_get_pred (q->z));

    while (tmp != q->z)
	{
	    if (tmp == _gdsl_node_get_succ (q->d))
		{
		    fprintf (file, "<GDSL_QUEUE_NODE REF=\"%p\" CONTENT=\"%p\" SUCC=\"%p\" PRED=\"\">", 
			     (void *) tmp, (void *) _gdsl_node_get_content (tmp), 
			     (void *) _gdsl_node_get_succ (tmp));
		}
	    else if (tmp == _gdsl_node_get_pred (q->z))
		{
		    fprintf (file, "<GDSL_QUEUE_NODE REF=\"%p\" CONTENT=\"%p\" SUCC=\"\" PRED=\"%p\">", 
			     (void *) tmp, (void *) _gdsl_node_get_content (tmp), 
			     (void *) _gdsl_node_get_pred (tmp));
		}
	    else
		{
		    fprintf (file, "<GDSL_QUEUE_NODE REF=\"%p\" CONTENT=\"%p\" SUCC=\"%p\" PRED=\"%p\">", 
			     (void *) tmp, (void *) _gdsl_node_get_content (tmp), 
			     (void *) _gdsl_node_get_succ (tmp), 
			     (void *) _gdsl_node_get_pred (tmp));
		}

	    if (write_f != NULL && _gdsl_node_get_content (tmp) != NULL)
		{
		    _gdsl_node_write (tmp, write_f, file, user_data);
		}

	    fprintf (file, "</GDSL_QUEUE_NODE>\n");

	    tmp = _gdsl_node_get_succ (tmp);
	}

    fprintf (file, "</GDSL_QUEUE>\n");
}

extern void
gdsl_queue_dump (const gdsl_queue_t q, gdsl_write_func_t write_f, FILE *file, 
		 void *user_data)
{
    _gdsl_node_t tmp;

    assert (q != NULL);
    assert (file != NULL);

    tmp = _gdsl_node_get_succ (q->d);

    fprintf (file, "<GDSL_QUEUE REF=\"%p\" NAME=\"%s\" CARD=\"%ld\" HEAD=\"%p\" TAIL=\"%p\">\n", 
	     (void *) q, q->name, q->card, (void *) q->d, (void *) q->z);

    fprintf (file, "<GDSL_QUEUE_HEAD REF=\"%p\" SUCC=\"%p\">\n", 
	     (void *) q->d, (void *) _gdsl_node_get_succ (q->d));

    while (tmp != q->z)
	{
	    if (_gdsl_node_get_content (tmp) != NULL)
		{
		    fprintf (file, "<GDSL_QUEUE_NODE REF=\"%p\" CONTENT=\"%p\" SUCC=\"%p\" PRED=\"%p\">", 
			     (void *) tmp, (void *) _gdsl_node_get_content (tmp), 
			     (void *) _gdsl_node_get_succ (tmp), 
			     (void *) _gdsl_node_get_pred (tmp));
		}
	    else
		{
		    fprintf (file, "<GDSL_QUEUE_NODE REF=\"%p\" CONTENT=\"\" SUCC=\"%p\" PRED=\"%p\">", 
			     (void *) tmp, (void *) _gdsl_node_get_succ (tmp), 
			     (void *) _gdsl_node_get_pred (tmp));
		}
      
	    if (write_f != NULL && _gdsl_node_get_content (tmp) != NULL)
		{
		    _gdsl_node_write (tmp, write_f, file, user_data);
		}
      
	    fprintf (file, "</GDSL_QUEUE_NODE>\n");

	    tmp = _gdsl_node_get_succ (tmp);
	}

    fprintf (file, "<GDSL_QUEUE_TAIL REF=\"%p\" PRED=\"%p\">\n</GDSL_QUEUE>\n", 
	     (void *) q->z, (void *) _gdsl_node_get_pred (q->z));
}

/******************************************************************************/
/* Private functions                                                          */
/******************************************************************************/

static gdsl_element_t
default_alloc (void* e)
{
    return e;
}

static void 
default_free (gdsl_element_t e)
{
    ;
}

static _gdsl_node_t
search_by_value (gdsl_queue_t q, gdsl_compare_func_t f, void* value)
{
    _gdsl_node_t left;
    _gdsl_node_t right;

    for (left = _gdsl_node_get_succ (q->d), right = _gdsl_node_get_pred (q->z);
	 left != _gdsl_node_get_succ (right);
	 left = _gdsl_node_get_succ (left), right = _gdsl_node_get_pred (right))
	{
	    if (f (_gdsl_node_get_content (left), value) == 0)
		{
		    return left;
		}
      
	    if (f (_gdsl_node_get_content (right), value) == 0)
		{
		    return right;
		}
      
	    if (left == right)
		{
		    return NULL;
		}
	}

    return NULL;
}

static _gdsl_node_t 
search_by_position (gdsl_queue_t q, ulong pos)
{
    ulong        m;
    _gdsl_node_t tmp;

    if (pos <= 0 || pos > q->card)
	{
	    return NULL;
	}

    m = (q->card / 2) + 1;

    if (pos < m)
	{
	    tmp = _gdsl_node_get_succ (q->d);

	    while (pos > 1)
		{
		    tmp = _gdsl_node_get_succ (tmp);
		    pos--;
		}
	}
    else
	{
	    pos = q->card - pos;
	    tmp = _gdsl_node_get_pred (q->z);

	    while (pos > 0)
		{
		    tmp = _gdsl_node_get_pred (tmp);
		    pos--;
		}
	}

    return tmp;
}

/** EMACS **
 * Local variables:
 * mode: c
 * c-basic-offset: 4
 * End:
 */
